@@ -1,24 +1,31 @@
 use {
     crate::{
-        ic_logger_msg, ic_msg, instruction_recorder::InstructionRecorder,
-        log_collector::LogCollector, native_loader::NativeLoader, pre_account::PreAccount,
-        timings::ExecuteDetailsTimings,
+        accounts_data_meter::AccountsDataMeter,
+        ic_logger_msg, ic_msg,
+        instruction_recorder::InstructionRecorder,
+        log_collector::LogCollector,
+        native_loader::NativeLoader,
+        pre_account::PreAccount,
+        timings::{ExecuteDetailsTimings, ExecuteTimings},
     },
+    solana_measure::measure::Measure,
     solana_sdk::{
         account::{AccountSharedData, ReadableAccount},
         account_utils::StateMut,
         bpf_loader_upgradeable::{self, UpgradeableLoaderState},
         compute_budget::ComputeBudget,
         feature_set::{
-            do_support_realloc, neon_evm_compute_budget, reject_empty_instruction_without_program,
-            remove_native_loader, requestable_heap_size, tx_wide_compute_cap, FeatureSet,
+            cap_accounts_data_len, do_support_realloc, neon_evm_compute_budget,
+            reject_empty_instruction_without_program, remove_native_loader, requestable_heap_size,
+            tx_wide_compute_cap, FeatureSet,
         },
         hash::Hash,
         instruction::{AccountMeta, CompiledInstruction, Instruction, InstructionError},
         keyed_account::{create_keyed_accounts_unified, keyed_account_at_index, KeyedAccount},
-        message::Message,
+        message::{Message, SanitizedMessage},
         pubkey::Pubkey,
         rent::Rent,
+        saturating_add_assign,
         sysvar::Sysvar,
     },
     std::{cell::RefCell, collections::HashMap, fmt::Debug, rc::Rc, sync::Arc},
@@ -30,6 +37,12 @@ pub type TransactionAccountRefCells = Vec<TransactionAccountRefCell>;
 pub type ProcessInstructionWithContext =
     fn(usize, &[u8], &mut InvokeContext) -> Result<(), InstructionError>;
 
+#[derive(Debug, PartialEq)]
+pub struct ProcessInstructionResult {
+    pub compute_units_consumed: u64,
+    pub result: Result<(), InstructionError>,
+}
+
 #[derive(Clone)]
 pub struct BuiltinProgram {
     pub program_id: Pubkey,
@@ -65,18 +78,57 @@ pub trait Executor: Debug + Send + Sync {
     ) -> Result<(), InstructionError>;
 }
 
-#[derive(Default)]
-pub struct Executors {
-    pub executors: HashMap<Pubkey, Arc<dyn Executor>>,
-    pub is_dirty: bool,
+pub type Executors = HashMap<Pubkey, TransactionExecutor>;
+
+/// Tracks whether a given executor is "dirty" and needs to updated in the
+/// executors cache
+pub struct TransactionExecutor {
+    executor: Arc<dyn Executor>,
+    is_miss: bool,
+    is_updated: bool,
 }
-impl Executors {
-    pub fn insert(&mut self, key: Pubkey, executor: Arc<dyn Executor>) {
-        let _ = self.executors.insert(key, executor);
-        self.is_dirty = true;
+
+impl TransactionExecutor {
+    /// Wraps an executor and tracks that it doesn't need to be updated in the
+    /// executors cache.
+    pub fn new_cached(executor: Arc<dyn Executor>) -> Self {
+        Self {
+            executor,
+            is_miss: false,
+            is_updated: false,
+        }
     }
-    pub fn get(&self, key: &Pubkey) -> Option<Arc<dyn Executor>> {
-        self.executors.get(key).cloned()
+
+    /// Wraps an executor and tracks that it needs to be updated in the
+    /// executors cache.
+    pub fn new_miss(executor: Arc<dyn Executor>) -> Self {
+        Self {
+            executor,
+            is_miss: true,
+            is_updated: false,
+        }
+    }
+
+    /// Wraps an executor and tracks that it needs to be updated in the
+    /// executors cache only if the transaction succeeded.
+    pub fn new_updated(executor: Arc<dyn Executor>) -> Self {
+        Self {
+            executor,
+            is_miss: false,
+            is_updated: true,
+        }
+    }
+
+    pub fn is_dirty(&self, include_updates: bool) -> bool {
+        self.is_miss || (include_updates && self.is_updated)
+    }
+
+    pub fn get(&self) -> Arc<dyn Executor> {
+        self.executor.clone()
+    }
+
+    pub fn clear_miss_for_test(&mut self) {
+        self.is_miss = false;
     }
 }
 
@@ -147,6 +199,7 @@ pub struct InvokeContext<'a> {
     compute_budget: ComputeBudget,
     current_compute_budget: ComputeBudget,
     compute_meter: Rc<RefCell<ComputeMeter>>,
+    accounts_data_meter: AccountsDataMeter,
     executors: Rc<RefCell<Executors>>,
     pub instruction_recorder: Option<&'a InstructionRecorder>,
     pub feature_set: Arc<FeatureSet>,
@@ -169,6 +222,7 @@ impl<'a> InvokeContext<'a> {
         feature_set: Arc<FeatureSet>,
         blockhash: Hash,
         lamports_per_signature: u64,
+        current_accounts_data_len: u64,
     ) -> Self {
         Self {
             invoke_stack: Vec::with_capacity(compute_budget.max_invoke_depth),
@@ -181,6 +235,7 @@ impl<'a> InvokeContext<'a> {
             current_compute_budget: compute_budget,
             compute_budget,
             compute_meter: ComputeMeter::new_ref(compute_budget.max_units),
+            accounts_data_meter: AccountsDataMeter::new(current_accounts_data_len),
             executors,
             instruction_recorder: None,
             feature_set,
@@ -206,13 +261,14 @@ impl<'a> InvokeContext<'a> {
             Arc::new(FeatureSet::all_enabled()),
             Hash::default(),
             0,
+            0,
         )
     }
 
     /// Push a stack frame onto the invocation stack
     pub fn push(
         &mut self,
-        message: &Message,
+        message: &SanitizedMessage,
         instruction: &CompiledInstruction,
         program_indices: &[usize],
         account_indices: &[usize],
@@ -327,12 +383,15 @@ impl<'a> InvokeContext<'a> {
     /// Verify the results of an instruction
     fn verify(
         &mut self,
-        message: &Message,
+        message: &SanitizedMessage,
         instruction: &CompiledInstruction,
         program_indices: &[usize],
     ) -> Result<(), InstructionError> {
-        let program_id = instruction.program_id(&message.account_keys);
+        let program_id = message
+            .get_account_key(instruction.program_id_index as usize)
+            .expect("invalid program id index");
         let do_support_realloc = self.feature_set.is_active(&do_support_realloc::id());
+        let cap_accounts_data_len = self.feature_set.is_active(&cap_accounts_data_len::id());
 
         // Verify all executable accounts have zero outstanding refs
         for account_index in program_indices.iter() {
@@ -379,6 +438,14 @@ impl<'a> InvokeContext<'a> {
             post_sum = post_sum
                 .checked_add(u128::from(account.lamports()))
                 .ok_or(InstructionError::UnbalancedInstruction)?;
+
+            if cap_accounts_data_len {
+                let pre_data_len = pre_account.data().len() as i64;
+                let post_data_len = account.data().len() as i64;
+                let data_len_delta = post_data_len.saturating_sub(pre_data_len);
+                self.accounts_data_meter.consume(data_len_delta)?;
+            }
+
             Ok(())
         };
         instruction.visit_each_account(&mut work)?;
@@ -398,6 +465,7 @@ impl<'a> InvokeContext<'a> {
         write_privileges: &[bool],
     ) -> Result<(), InstructionError> {
         let do_support_realloc = self.feature_set.is_active(&do_support_realloc::id());
+        let cap_accounts_data_len = self.feature_set.is_active(&cap_accounts_data_len::id());
         let program_id = self
             .invoke_stack
             .last()
@@ -456,6 +524,14 @@ impl<'a> InvokeContext<'a> {
                         if is_writable && !pre_account.executable() {
                             pre_account.update(&account);
                         }
+
+                        if cap_accounts_data_len {
+                            let pre_data_len = pre_account.data().len() as i64;
+                            let post_data_len = account.data().len() as i64;
+                            let data_len_delta = post_data_len.saturating_sub(pre_data_len);
+                            self.accounts_data_meter.consume(data_len_delta)?;
+                        }
+
                         return Ok(());
                     }
                 }
@@ -493,13 +569,17 @@ impl<'a> InvokeContext<'a> {
         if let Some(instruction_recorder) = &self.instruction_recorder {
             instruction_recorder.record_instruction(instruction);
         }
+
+        let message = SanitizedMessage::Legacy(message);
         self.process_instruction(
             &message,
-            &message.instructions[0],
+            &message.instructions()[0],
             &program_indices,
             &account_indices,
             &caller_write_privileges,
-        )?;
+            &mut ExecuteTimings::default(),
+        )
+        .result?;
 
         // Verify the called program has not misbehaved
         let do_support_realloc = self.feature_set.is_active(&do_support_realloc::id());
@@ -635,42 +715,89 @@ impl<'a> InvokeContext<'a> {
     /// Processes a cross-program instruction and returns how many compute units were used
     pub fn process_instruction(
         &mut self,
-        message: &Message,
+        message: &SanitizedMessage,
         instruction: &CompiledInstruction,
         program_indices: &[usize],
         account_indices: &[usize],
         caller_write_privileges: &[bool],
-    ) -> Result<u64, InstructionError> {
+        timings: &mut ExecuteTimings,
+    ) -> ProcessInstructionResult {
         let is_lowest_invocation_level = self.invoke_stack.is_empty();
         if !is_lowest_invocation_level {
             // Verify the calling program hasn't misbehaved
-            self.verify_and_update(instruction, account_indices, caller_write_privileges)?;
+            let mut verify_caller_time = Measure::start("verify_caller_time");
+            let result =
+                self.verify_and_update(instruction, account_indices, caller_write_privileges);
+            verify_caller_time.stop();
+            saturating_add_assign!(
+                timings
+                    .execute_accessories
+                    .process_instructions
+                    .verify_caller_us,
+                verify_caller_time.as_us()
+            );
+            if result.is_err() {
+                return ProcessInstructionResult {
+                    compute_units_consumed: 0,
+                    result,
+                };
+            }
         }
 
+        let mut compute_units_consumed = 0;
         let result = self
             .push(message, instruction, program_indices, account_indices)
             .and_then(|_| {
-                self.return_data = (*instruction.program_id(&message.account_keys), Vec::new());
+                let mut process_executable_chain_time =
+                    Measure::start("process_executable_chain_time");
+                let program_id = message
+                    .get_account_key(instruction.program_id_index as usize)
+                    .expect("invalid program id index");
+                self.return_data = (*program_id, Vec::new());
                 let pre_remaining_units = self.compute_meter.borrow().get_remaining();
-                self.process_executable_chain(&instruction.data)?;
+                let execution_result = self.process_executable_chain(&instruction.data);
                 let post_remaining_units = self.compute_meter.borrow().get_remaining();
+                compute_units_consumed = pre_remaining_units.saturating_sub(post_remaining_units);
+                process_executable_chain_time.stop();
 
                 // Verify the called program has not misbehaved
-                if is_lowest_invocation_level {
-                    self.verify(message, instruction, program_indices)?;
-                } else {
-                    let write_privileges: Vec<bool> = (0..message.account_keys.len())
-                        .map(|i| message.is_writable(i))
-                        .collect();
-                    self.verify_and_update(instruction, account_indices, &write_privileges)?;
-                }
+                let mut verify_callee_time = Measure::start("verify_callee_time");
+                let result = execution_result.and_then(|_| {
+                    if is_lowest_invocation_level {
+                        self.verify(message, instruction, program_indices)
+                    } else {
+                        let write_privileges: Vec<bool> = (0..message.account_keys_len())
+                            .map(|i| message.is_writable(i))
+                            .collect();
+                        self.verify_and_update(instruction, account_indices, &write_privileges)
+                    }
+                });
+                verify_callee_time.stop();
 
-                Ok(pre_remaining_units.saturating_sub(post_remaining_units))
+                saturating_add_assign!(
+                    timings
+                        .execute_accessories
+                        .process_instructions
+                        .process_executable_chain_us,
+                    process_executable_chain_time.as_us()
+                );
+                saturating_add_assign!(
+                    timings
+                        .execute_accessories
+                        .process_instructions
+                        .verify_callee_us,
+                    verify_callee_time.as_us()
+                );
+
+                result
             });
 
         // Pop the invoke_stack to restore previous state
         self.pop();
-        result
+        ProcessInstructionResult {
+            compute_units_consumed,
+            result,
+        }
     }
 
     /// Calls the instruction's program entrypoint method
@@ -781,15 +908,31 @@ impl<'a> InvokeContext<'a> {
         self.compute_meter.clone()
     }
 
-    /// Loaders may need to do work in order to execute a program. Cache
-    /// the work that can be re-used across executions
+    /// Get this invocation's AccountsDataMeter
+    pub fn get_accounts_data_meter(&self) -> &AccountsDataMeter {
+        &self.accounts_data_meter
+    }
+
+    /// Cache an executor that wasn't found in the cache
     pub fn add_executor(&self, pubkey: &Pubkey, executor: Arc<dyn Executor>) {
-        self.executors.borrow_mut().insert(*pubkey, executor);
+        self.executors
+            .borrow_mut()
+            .insert(*pubkey, TransactionExecutor::new_miss(executor));
+    }
+
+    /// Cache an executor that has changed
+    pub fn update_executor(&self, pubkey: &Pubkey, executor: Arc<dyn Executor>) {
+        self.executors
+            .borrow_mut()
+            .insert(*pubkey, TransactionExecutor::new_updated(executor));
     }
 
     /// Get the completed loader work that can be re-used across execution
     pub fn get_executor(&self, pubkey: &Pubkey) -> Option<Arc<dyn Executor>> {
-        self.executors.borrow().get(pubkey)
+        self.executors
+            .borrow()
+            .get(pubkey)
+            .map(|tx_executor| tx_executor.executor.clone())
     }
 
     /// Find an account_index and account by its key
@@ -827,7 +970,7 @@ impl<'a> InvokeContext<'a> {
 
 pub struct MockInvokeContextPreparation {
     pub accounts: TransactionAccountRefCells,
-    pub message: Message,
+    pub message: SanitizedMessage,
     pub account_indices: Vec<usize>,
 }
 
@@ -858,17 +1001,16 @@ pub fn prepare_mock_invoke_context(
     for program_index in program_indices.iter().rev() {
         metas.remove(*program_index);
     }
-    let message = Message::new(
+    let message = SanitizedMessage::Legacy(Message::new(
         &[Instruction::new_with_bytes(
             program_id,
             instruction_data,
             metas,
         )],
         None,
-    );
+    ));
     let account_indices: Vec<usize> = message
-        .account_keys
-        .iter()
+        .account_keys_iter()
         .map(|search_key| {
             accounts
                 .iter()
@@ -914,7 +1056,7 @@ pub fn with_mock_invoke_context<R, F: FnMut(&mut InvokeContext) -> R>(
     invoke_context
         .push(
             &preparation.message,
-            &preparation.message.instructions[0],
+            &preparation.message.instructions()[0],
             &program_indices,
             &preparation.account_indices,
         )
@@ -939,7 +1081,7 @@ pub fn mock_process_instruction_with_sysvars(
     invoke_context.sysvars = sysvars;
     invoke_context.push(
         &preparation.message,
-        &preparation.message.instructions[0],
+        &preparation.message.instructions()[0],
         &program_indices,
         &preparation.account_indices,
     )?;
@@ -983,6 +1125,13 @@ mod tests {
         ModifyOwned,
         ModifyNotOwned,
         ModifyReadonly,
+        ConsumeComputeUnits {
+            compute_units_consumed: u64,
+            desired_result: Result<(), InstructionError>,
+        },
+        Resize {
+            new_len: usize,
+        },
     }
 
     #[test]
@@ -1052,6 +1201,23 @@ mod tests {
                         .try_account_ref_mut()?
                         .data_as_mut_slice()[0] = 1
                 }
+                MockInstruction::ConsumeComputeUnits {
+                    compute_units_consumed,
+                    desired_result,
+                } => {
+                    invoke_context
+                        .get_compute_meter()
+                        .borrow_mut()
+                        .consume(compute_units_consumed)
+                        .unwrap();
+                    return desired_result;
+                }
+                MockInstruction::Resize { new_len } => {
+                    keyed_account_at_index(keyed_accounts, first_instruction_account)?
+                        .try_account_ref_mut()?
+                        .data_mut()
+                        .resize_with(new_len, Default::default)
+                }
             }
         } else {
             return Err(InstructionError::InvalidInstructionData);
@@ -1090,10 +1256,10 @@ mod tests {
         }
         let account_indices = (0..accounts.len()).collect::<Vec<usize>>();
 
-        let message = Message::new(
+        let message = SanitizedMessage::Legacy(Message::new(
             &[Instruction::new_with_bytes(invoke_stack[0], &[0], metas)],
             None,
-        );
+        ));
         let mut invoke_context = InvokeContext::new_mock(&accounts, &[]);
 
         // Check call depth increases and has a limit
@@ -1102,7 +1268,7 @@ mod tests {
             if Err(InstructionError::CallDepth)
                 == invoke_context.push(
                     &message,
-                    &message.instructions[0],
+                    &message.instructions()[0],
                     &[MAX_DEPTH + depth_reached],
                     &[],
                 )
@@ -1173,25 +1339,25 @@ mod tests {
             solana_sdk::pubkey::new_rand(),
             Rc::new(RefCell::new(AccountSharedData::default())),
         )];
-        let message = Message::new(
+        let message = SanitizedMessage::Legacy(Message::new(
             &[Instruction::new_with_bincode(
                 accounts[0].0,
                 &MockInstruction::NoopSuccess,
                 vec![AccountMeta::new_readonly(accounts[0].0, false)],
             )],
             None,
-        );
+        ));
         let mut invoke_context = InvokeContext::new_mock(&accounts, &[]);
         invoke_context
-            .push(&message, &message.instructions[0], &[0], &[])
+            .push(&message, &message.instructions()[0], &[0], &[])
             .unwrap();
         assert!(invoke_context
-            .verify(&message, &message.instructions[0], &[0])
+            .verify(&message, &message.instructions()[0], &[0])
             .is_ok());
 
         let mut _borrowed = accounts[0].1.borrow();
         assert_eq!(
-            invoke_context.verify(&message, &message.instructions[0], &[0]),
+            invoke_context.verify(&message, &message.instructions()[0], &[0]),
             Err(InstructionError::AccountBorrowOutstanding)
         );
     }
@@ -1240,7 +1406,7 @@ mod tests {
             &MockInstruction::NoopSuccess,
             metas.clone(),
         );
-        let message = Message::new(&[callee_instruction], None);
+        let message = SanitizedMessage::Legacy(Message::new(&[callee_instruction], None));
 
         let builtin_programs = &[BuiltinProgram {
             program_id: callee_program_id,
@@ -1253,20 +1419,22 @@ mod tests {
 
         // not owned account modified by the caller (before the invoke)
         let caller_write_privileges = message
-            .account_keys
-            .iter()
+            .account_keys_iter()
             .enumerate()
             .map(|(i, _)| message.is_writable(i))
             .collect::<Vec<bool>>();
         accounts[0].1.borrow_mut().data_as_mut_slice()[0] = 1;
         assert_eq!(
-            invoke_context.process_instruction(
-                &message,
-                &message.instructions[0],
-                &program_indices[1..],
-                &account_indices,
-                &caller_write_privileges,
-            ),
+            invoke_context
+                .process_instruction(
+                    &message,
+                    &message.instructions()[0],
+                    &program_indices[1..],
+                    &account_indices,
+                    &caller_write_privileges,
+                    &mut ExecuteTimings::default(),
+                )
+                .result,
             Err(InstructionError::ExternalAccountDataModified)
         );
         accounts[0].1.borrow_mut().data_as_mut_slice()[0] = 0;
@@ -1274,13 +1442,16 @@ mod tests {
         // readonly account modified by the invoker
         accounts[2].1.borrow_mut().data_as_mut_slice()[0] = 1;
         assert_eq!(
-            invoke_context.process_instruction(
-                &message,
-                &message.instructions[0],
-                &program_indices[1..],
-                &account_indices,
-                &caller_write_privileges,
-            ),
+            invoke_context
+                .process_instruction(
+                    &message,
+                    &message.instructions()[0],
+                    &program_indices[1..],
+                    &account_indices,
+                    &caller_write_privileges,
+                    &mut ExecuteTimings::default(),
+                )
+                .result,
             Err(InstructionError::ReadonlyDataModified)
         );
         accounts[2].1.borrow_mut().data_as_mut_slice()[0] = 0;
@@ -1288,37 +1459,55 @@ mod tests {
         invoke_context.pop();
 
         let cases = vec![
-            (MockInstruction::NoopSuccess, Ok(0)),
+            (
+                MockInstruction::NoopSuccess,
+                ProcessInstructionResult {
+                    result: Ok(()),
+                    compute_units_consumed: 0,
+                },
+            ),
             (
                 MockInstruction::NoopFail,
-                Err(InstructionError::GenericError),
+                ProcessInstructionResult {
+                    result: Err(InstructionError::GenericError),
+                    compute_units_consumed: 0,
+                },
+            ),
+            (
+                MockInstruction::ModifyOwned,
+                ProcessInstructionResult {
+                    result: Ok(()),
+                    compute_units_consumed: 0,
+                },
             ),
-            (MockInstruction::ModifyOwned, Ok(0)),
             (
                 MockInstruction::ModifyNotOwned,
-                Err(InstructionError::ExternalAccountDataModified),
+                ProcessInstructionResult {
+                    result: Err(InstructionError::ExternalAccountDataModified),
+                    compute_units_consumed: 0,
+                },
             ),
         ];
         for case in cases {
             let callee_instruction =
                 Instruction::new_with_bincode(callee_program_id, &case.0, metas.clone());
-            let message = Message::new(&[callee_instruction], None);
+            let message = SanitizedMessage::Legacy(Message::new(&[callee_instruction], None));
             invoke_context
                 .push(&message, &caller_instruction, &program_indices[..1], &[])
                 .unwrap();
             let caller_write_privileges = message
-                .account_keys
-                .iter()
+                .account_keys_iter()
                 .enumerate()
                 .map(|(i, _)| message.is_writable(i))
                 .collect::<Vec<bool>>();
             assert_eq!(
                 invoke_context.process_instruction(
                     &message,
-                    &message.instructions[0],
+                    &message.instructions()[0],
                     &program_indices[1..],
                     &account_indices,
                     &caller_write_privileges,
+                    &mut ExecuteTimings::default(),
                 ),
                 case.1
             );
@@ -1368,7 +1557,7 @@ mod tests {
             &MockInstruction::NoopSuccess,
             metas.clone(),
         );
-        let message = Message::new(&[callee_instruction.clone()], None);
+        let message = SanitizedMessage::Legacy(Message::new(&[callee_instruction.clone()], None));
 
         let builtin_programs = &[BuiltinProgram {
             program_id: callee_program_id,
@@ -1417,7 +1606,8 @@ mod tests {
         for case in cases {
             let callee_instruction =
                 Instruction::new_with_bincode(callee_program_id, &case.0, metas.clone());
-            let message = Message::new(&[callee_instruction.clone()], None);
+            let message =
+                SanitizedMessage::Legacy(Message::new(&[callee_instruction.clone()], None));
             invoke_context
                 .push(&message, &caller_instruction, &program_indices, &[])
                 .unwrap();
@@ -1442,22 +1632,22 @@ mod tests {
             ),
         ];
 
-        let noop_message = Message::new(
+        let noop_message = SanitizedMessage::Legacy(Message::new(
             &[Instruction::new_with_bincode(
                 accounts[0].0,
                 &MockInstruction::NoopSuccess,
                 vec![AccountMeta::new_readonly(accounts[0].0, false)],
             )],
             None,
-        );
-        let neon_message = Message::new(
+        ));
+        let neon_message = SanitizedMessage::Legacy(Message::new(
             &[Instruction::new_with_bincode(
                 crate::neon_evm_program::id(),
                 &MockInstruction::NoopSuccess,
                 vec![AccountMeta::new_readonly(accounts[0].0, false)],
             )],
             None,
-        );
+        ));
 
         let mut feature_set = FeatureSet::all_enabled();
         feature_set.deactivate(&tx_wide_compute_cap::id());
@@ -1466,7 +1656,7 @@ mod tests {
         invoke_context.feature_set = Arc::new(feature_set);
 
         invoke_context
-            .push(&noop_message, &noop_message.instructions[0], &[0], &[])
+            .push(&noop_message, &noop_message.instructions()[0], &[0], &[])
             .unwrap();
         assert_eq!(
             *invoke_context.get_compute_budget(),
@@ -1475,7 +1665,7 @@ mod tests {
         invoke_context.pop();
 
         invoke_context
-            .push(&neon_message, &neon_message.instructions[0], &[1], &[])
+            .push(&neon_message, &neon_message.instructions()[0], &[1], &[])
             .unwrap();
         let expected_compute_budget = ComputeBudget {
             max_units: 500_000,
@@ -1489,7 +1679,7 @@ mod tests {
         invoke_context.pop();
 
         invoke_context
-            .push(&noop_message, &noop_message.instructions[0], &[0], &[])
+            .push(&noop_message, &noop_message.instructions()[0], &[0], &[])
             .unwrap();
         assert_eq!(
             *invoke_context.get_compute_budget(),
@@ -1497,4 +1687,223 @@ mod tests {
         );
         invoke_context.pop();
     }
+
+    #[test]
+    fn test_process_instruction_compute_budget() {
+        let caller_program_id = solana_sdk::pubkey::new_rand();
+        let callee_program_id = solana_sdk::pubkey::new_rand();
+        let owned_account = AccountSharedData::new(42, 1, &callee_program_id);
+        let not_owned_account = AccountSharedData::new(84, 1, &solana_sdk::pubkey::new_rand());
+        let readonly_account = AccountSharedData::new(168, 1, &solana_sdk::pubkey::new_rand());
+        let loader_account = AccountSharedData::new(0, 0, &native_loader::id());
+        let mut program_account = AccountSharedData::new(1, 0, &native_loader::id());
+        program_account.set_executable(true);
+
+        let accounts = vec![
+            (
+                solana_sdk::pubkey::new_rand(),
+                Rc::new(RefCell::new(owned_account)),
+            ),
+            (
+                solana_sdk::pubkey::new_rand(),
+                Rc::new(RefCell::new(not_owned_account)),
+            ),
+            (
+                solana_sdk::pubkey::new_rand(),
+                Rc::new(RefCell::new(readonly_account)),
+            ),
+            (caller_program_id, Rc::new(RefCell::new(loader_account))),
+            (callee_program_id, Rc::new(RefCell::new(program_account))),
+        ];
+        let account_indices = [0, 1, 2];
+        let program_indices = [3, 4];
+
+        let metas = vec![
+            AccountMeta::new(accounts[0].0, false),
+            AccountMeta::new(accounts[1].0, false),
+            AccountMeta::new_readonly(accounts[2].0, false),
+        ];
+
+        let builtin_programs = &[BuiltinProgram {
+            program_id: callee_program_id,
+            process_instruction: mock_process_instruction,
+        }];
+        let mut invoke_context = InvokeContext::new_mock(&accounts, builtin_programs);
+
+        let compute_units_consumed = 10;
+        let desired_results = vec![Ok(()), Err(InstructionError::GenericError)];
+
+        for desired_result in desired_results {
+            let caller_instruction =
+                CompiledInstruction::new(program_indices[0] as u8, &(), vec![0, 1, 2, 3, 4]);
+            let callee_instruction = Instruction::new_with_bincode(
+                callee_program_id,
+                &MockInstruction::ConsumeComputeUnits {
+                    compute_units_consumed,
+                    desired_result: desired_result.clone(),
+                },
+                metas.clone(),
+            );
+            let message =
+                SanitizedMessage::Legacy(Message::new(&[callee_instruction.clone()], None));
+            invoke_context
+                .push(&message, &caller_instruction, &program_indices[..1], &[])
+                .unwrap();
+            let caller_write_privileges = message
+                .account_keys_iter()
+                .enumerate()
+                .map(|(i, _)| message.is_writable(i))
+                .collect::<Vec<bool>>();
+            let result = invoke_context.process_instruction(
+                &message,
+                &message.instructions()[0],
+                &program_indices[1..],
+                &account_indices,
+                &caller_write_privileges,
+                &mut ExecuteTimings::default(),
+            );
+
+            // Because the instruction had compute cost > 0, then regardless of the execution result,
+            // the number of compute units consumed should be a non-default which is something greater
+            // than zero.
+            assert!(result.compute_units_consumed > 0);
+            assert_eq!(
+                result,
+                ProcessInstructionResult {
+                    compute_units_consumed,
+                    result: desired_result,
+                }
+            );
+        }
+    }
+
+    #[test]
+    fn test_process_instruction_accounts_data_meter() {
+        let program_key = Pubkey::new_unique();
+        let user_account_data_len = 123;
+        let user_account = AccountSharedData::new(100, user_account_data_len, &program_key);
+        let dummy1_account = AccountSharedData::new(10, 0, &program_key);
+        let mut program_account = AccountSharedData::new(500, 500, &native_loader::id());
+        program_account.set_executable(true);
+        let accounts = vec![
+            (Pubkey::new_unique(), Rc::new(RefCell::new(user_account))),
+            (Pubkey::new_unique(), Rc::new(RefCell::new(dummy1_account))),
+            (program_key, Rc::new(RefCell::new(program_account))),
+        ];
+        let account_indices = [];
+        let program_indices = [accounts.len() - 1];
+
+        let metas = accounts
+            .iter()
+            .map(|account| AccountMeta::new(account.0, false))
+            .collect::<Vec<_>>();
+
+        let builtin_programs = [BuiltinProgram {
+            program_id: program_key,
+            process_instruction: mock_process_instruction,
+        }];
+
+        let mut invoke_context = InvokeContext::new_mock(&accounts, &builtin_programs);
+        invoke_context
+            .accounts_data_meter
+            .set_current(user_account_data_len as u64);
+        invoke_context
+            .accounts_data_meter
+            .set_maximum(user_account_data_len as u64 * 3);
+        let remaining_account_data_len = invoke_context.accounts_data_meter.remaining() as usize;
+
+        // Test 1: Resize the account to use up all the space; this must succeed
+        {
+            let new_len = user_account_data_len + remaining_account_data_len;
+            let instruction = Instruction::new_with_bincode(
+                program_key,
+                &MockInstruction::Resize { new_len },
+                metas.clone(),
+            );
+            let message = SanitizedMessage::Legacy(Message::new(&[instruction], None));
+            let caller_write_privileges = message
+                .account_keys_iter()
+                .enumerate()
+                .map(|(i, _)| message.is_writable(i))
+                .collect::<Vec<_>>();
+
+            let result = invoke_context
+                .process_instruction(
+                    &message,
+                    &message.instructions()[0],
+                    &program_indices,
+                    &account_indices,
+                    &caller_write_privileges,
+                    &mut ExecuteTimings::default(),
+                )
+                .result;
+
+            assert!(result.is_ok());
+            assert_eq!(invoke_context.accounts_data_meter.remaining(), 0);
+        }
+
+        // Test 2: Resize the account to *the same size*, so not consuming any additional size; this must succeed
+        {
+            let new_len = user_account_data_len + remaining_account_data_len;
+            let instruction = Instruction::new_with_bincode(
+                program_key,
+                &MockInstruction::Resize { new_len },
+                metas.clone(),
+            );
+            let message = SanitizedMessage::Legacy(Message::new(&[instruction], None));
+            let caller_write_privileges = message
+                .account_keys_iter()
+                .enumerate()
+                .map(|(i, _)| message.is_writable(i))
+                .collect::<Vec<_>>();
+
+            let result = invoke_context
+                .process_instruction(
+                    &message,
+                    &message.instructions()[0],
+                    &program_indices,
+                    &account_indices,
+                    &caller_write_privileges,
+                    &mut ExecuteTimings::default(),
+                )
+                .result;
+
+            assert!(result.is_ok());
+            assert_eq!(invoke_context.accounts_data_meter.remaining(), 0);
+        }
+
+        // Test 3: Resize the account to exceed the budget; this must fail
+        {
+            let new_len = user_account_data_len + remaining_account_data_len + 1;
+            let instruction = Instruction::new_with_bincode(
+                program_key,
+                &MockInstruction::Resize { new_len },
+                metas,
+            );
+            let message = SanitizedMessage::Legacy(Message::new(&[instruction], None));
+            let caller_write_privileges = message
+                .account_keys_iter()
+                .enumerate()
+                .map(|(i, _)| message.is_writable(i))
+                .collect::<Vec<_>>();
+
+            let result = invoke_context
+                .process_instruction(
+                    &message,
+                    &message.instructions()[0],
+                    &program_indices,
+                    &account_indices,
+                    &caller_write_privileges,
+                    &mut ExecuteTimings::default(),
+                )
+                .result;
+
+            assert!(result.is_err());
+            assert!(matches!(
+                result,
+                Err(solana_sdk::instruction::InstructionError::AccountsDataBudgetExceeded)
+            ));
+            assert_eq!(invoke_context.accounts_data_meter.remaining(), 0);
+        }
+    }
 }
